import copy
from typing import Iterable
import torch

from modelbest_sdk.dataset.thrift_wrapper.dataset_context import DatasetContext


class CudaPrefetcher(Iterable):
    """
    Wrap around a batch iterator for asynchornously copying data to gpu to shield memcpy latency.
    """

    def __init__(self, context: DatasetContext, loader):
        self.context = context
        self.loader = iter(loader)
        self.stream = torch.cuda.Stream()
        self.preload()

    def preload(self):
        try:
            self.data = next(self.loader)
        except StopIteration:
            self.data = None
            return
        with torch.cuda.stream(self.stream):
            for key in self.data.keys():
                if isinstance(self.data[key], torch.Tensor):
                    self.data[key] = self.data[key].cuda(non_blocking=True)

    def __next__(self):
        torch.cuda.current_stream().wait_stream(self.stream)
        data = copy.deepcopy(self.data)
        self.preload()
        return data

    def __iter__(self):
        return self



import unittest

from modelbest_sdk.dataset.thrift_wrapper.dataset_context import DatasetContext


class TestDatasetContext(unittest.TestCase):
    def test_read_from_file(self):
        simple_dataset_context = DatasetContext(
            rank=1,
            world_size=10,
            dataset_config_path="/path/to/dataset_config",
            dataset_checkpoint_path="/path/to/checkpoint",
        )
        simple_dataset_context.save_to_file("/tmp/test_context.mbt")
        new_context = DatasetContext.load_from_file("/tmp/test_context.mbt")
        assert new_context.rank == simple_dataset_context.rank
        assert new_context.world_size == simple_dataset_context.world_size
        assert new_context.dataset_config_path == simple_dataset_context.dataset_config_path
        assert new_context.dataset_checkpoint_path == simple_dataset_context.dataset_checkpoint_path



import os
import torch
from modelbest_sdk.dataset.batched_dataset import BatchedDataset
from modelbest_sdk.dataset.megatron.segment_dataset import SegmentDataset
from modelbest_sdk.dataset.thrift_wrapper.dataset_checkpoint import DatasetCheckpointList, DatasetInfoList
from modelbest_sdk.dataset.thrift_wrapper.dataset_context import DatasetContext
from modelbest_sdk.dataset.weighted_dataset import WeightedDataset

class ModelbestDataloader():
    def __init__(
        self,
        context: DatasetContext,
        dataset_info_list: DatasetInfoList,
        batch_size=1,
        max_len=4096,
        prefetch_chunk_cnt=20,
        chunk_size=1024,
        num_workers=1,
        prefetch_factor=20,
        cuda_prefetch=True
    ):
        self.context = context
        self.context.num_workers = num_workers
        
        self.weighted_dataset = WeightedDataset(
            context=context, 
            dataset_info_list=dataset_info_list,
            prefetch_chunk_cnt=prefetch_chunk_cnt,
            chunk_size=chunk_size
        )
        self.setup_dataset(context, batch_size, max_len)
        
        self.dataloader = torch.utils.data.DataLoader(
            dataset=self.batched_dataset, 
            batch_size=batch_size,
            num_workers=num_workers,
            prefetch_factor=prefetch_factor,
            pin_memory=True,
            collate_fn=self.dataset_collate_fn,
        )
        
        self.cuda_prefetch = cuda_prefetch
        
        if self.cuda_prefetch:
            from modelbest_sdk.dataset.cuda_prefetcher import CudaPrefetcher
            self.cuda_prefetcher = CudaPrefetcher(
                context,
                self.dataloader
            )
        self.iterator = None

    def setup_dataset(self, context, batch_size, max_len):
        raise NotImplementedError("This method should be overridden by subclasses.")

    def dataset_collate_fn(self):
        raise NotImplementedError("This method should be overridden by subclasses.")
    
    def __iter__(self):
        if not self.cuda_prefetch:
            self.iterator = iter(self.dataloader)
        else:
            self.iterator = iter(self.cuda_prefetcher)
        return self
    
    def __next__(self):
        if self.iterator is None:
            self.__iter__()
        return next(self.iterator)
    
    def checkpoint(self):
        return self.weighted_dataset.checkpoint()
    
    def load_checkpoint(self, checkpoint):
        self.weighted_dataset.load_checkpoint(checkpoint)
        
    def update(self, dataset_entries):
        self.weighted_dataset.update(dataset_entries)
    
    def save(self, sub_dir=''):
        path = os.path.join(self.context.dataset_checkpoint_path, sub_dir, f"dataset_ckpt_rank_{self.context.rank}.mbt")
        self.checkpoint().save_to_file(path)
        
    def resume(self, sub_dir=''):
        # if dir empty, return
        resume_dir = os.path.join(self.context.dataset_checkpoint_path, sub_dir)
        if not os.path.exists(resume_dir):
            return
        if os.listdir(resume_dir) == []:
            return
        new_world_size = self.context.world_size
        first_rank_ckpt_path = os.path.join(resume_dir, f"dataset_ckpt_rank_0.mbt")
        cur_rank_ckpt_path = os.path.join(resume_dir, f"dataset_ckpt_rank_{self.context.rank}.mbt")
        old_world_size = DatasetCheckpointList.load_from_file(first_rank_ckpt_path).world_size
        if new_world_size == old_world_size:
            self.weighted_dataset.load_checkpoint(DatasetCheckpointList.load_from_file(cur_rank_ckpt_path))
        else:
            print(f"Warning: world size changed from {old_world_size} to {new_world_size}, resuming from scratch")
            merged_ckpt = None
            for path in os.listdir(resume_dir):
                abs_path = os.path.join(resume_dir, path)
                if merged_ckpt is None:
                    merged_ckpt = DatasetCheckpointList.load_from_file(abs_path)
                else:
                    merged_ckpt.merge(DatasetCheckpointList.load_from_file(abs_path))
            self.load_checkpoint(merged_ckpt)
            
class MegatronExternalDataloader(ModelbestDataloader):
    def setup_dataset(self, context, batch_size, max_len):
        self.batched_dataset = SegmentDataset(
            context=context,
            weighted_dataset=self.weighted_dataset,
            max_len=max_len
        )
        
    @staticmethod
    def dataset_collate_fn(batch):
        return SegmentDataset.collate_fn(batch)

class CpmFlashAttnDataloader(ModelbestDataloader):
    def setup_dataset(self, context, batch_size, max_len):
        self.batched_dataset = BatchedDataset(
            context=context, 
            weighted_dataset=self.weighted_dataset, 
            batch_size=batch_size, 
            max_len=max_len)
    
    @staticmethod
    def dataset_collate_fn(batch):
        return BatchedDataset.collate_fn(batch)



import numpy as np
from modelbest_sdk.dataset.sampler.sampler import Sampler


class WeightedSampler(Sampler):
    def __init__(self, weights, seed):
        self.weights = weights
        self.n_samples = len(weights)
        np.random.seed(seed)
        
    def __call__(self):
        return np.random.choice(self.n_samples, p=self.weights)


from collections import defaultdict
import hashlib
import numpy as np
import torch

# from modelbest_sdk.dataset.collater.collater import Collater
from modelbest_sdk.dataset.weighted_dataset import WeightedDataset
from modelbest_sdk.dataset.thrift_wrapper.base_doc import BaseDoc
from modelbest_sdk.dataset.thrift_wrapper.dataset_checkpoint import *
from modelbest_sdk.dataset.thrift_wrapper.dataset_context import DatasetContext

def default_factory_list():
    return list()

def default_factory_dict():
    return defaultdict(default_factory_list)
class BatchedDataset(torch.utils.data.IterableDataset):
    def __init__(self, context, weighted_dataset, batch_size, max_len, drop_last=False):
        self.context = context
        self.weighted_dataset = weighted_dataset
        self.max_total_length = batch_size * max_len
        self.batch_size = 1
        self.drop_last = drop_last
        
        self.buffer = []
        self.current_length = 0

        # 初始化tag_to_index
        self.tag_to_hash = defaultdict()
      
    def put(self, data):
        self.buffer.append(data)
        self.current_length += len(data['doc'].token_ids)
    
    def pop(self):
        lengths = []
        indexes = defaultdict(default_factory_dict)        
        inputs = torch.zeros((self.batch_size, self.max_total_length), dtype=torch.int32)
        targets = torch.full((self.batch_size, self.max_total_length), dtype=torch.int32, fill_value=-100)
        dataset_ids = torch.full((self.batch_size, self.max_total_length), dtype=torch.int32, fill_value=-1)
        position_ids = torch.zeros((self.batch_size, self.max_total_length), dtype=torch.int32)
        tags = torch.full((self.batch_size, self.max_total_length), dtype=torch.int64, fill_value=-1)

        span_begin = 0
        while self.buffer:
            data = self.buffer.pop()
            chunk = data['chunk']
            index = data['index']
            dataset_idx = data['dataset_idx']
            doc: BaseDoc = data['doc']
            token_ids = doc.token_ids
            mask = doc.mask
            tag = doc.tag[0] # TODO: support multiple tags
            target_ids = np.where(mask[1:], -100, token_ids[1:]).tolist() + [-100]
            span_end = span_begin + len(token_ids)
            # TODO: what if the inputs is longer than max_total_length?
            # RuntimeError: The expanded size of the tensor (16) must match the existing size (385) at non-singleton dimension 0.  Target sizes: [16].  Tensor sizes: [385]
            inputs[0, span_begin:span_end] = torch.tensor(token_ids, dtype=torch.int32)
            targets[0, span_begin:span_end] = torch.tensor(target_ids, dtype=torch.int32)
            dataset_ids[0, span_begin:span_end] = torch.tensor(dataset_idx, dtype=torch.int32)
            position_ids[0, span_begin:span_end] = torch.from_numpy(np.arange(len(token_ids), dtype=np.int32))
            lengths.append(len(token_ids))
            indexes[int(dataset_idx)][chunk].append(index)
            tags[0, span_begin:span_end] = self.encode_tags(token_ids, tag)
            span_begin = span_end
        cu_seqlens = torch.cat(
            [torch.tensor([0] + lengths).cumsum(dim=-1), torch.tensor([self.max_total_length], dtype=torch.int32)],
            dim=0,
        ).int()
        batch = {
            "input_ids": inputs,
            "target_ids": targets,
            "dataset_ids": dataset_ids,
            "indexes": indexes,
            "cu_seqlens": cu_seqlens,
            "max_seqlen": int(torch.max(cu_seqlens[1:] - cu_seqlens[:-1])),
            "lengths": torch.tensor(sum(lengths)).int(),
            "position_ids": position_ids,
            "tags": tags,
            "tag_to_hash": self.tag_to_hash
        }
        self.current_length = 0
        return batch
        

    def will_exceed(self, data):
        return self.current_length + len(data['doc'].token_ids) > self.max_total_length

    def __iter__(self):
        for data in self.weighted_dataset:
            if self.will_exceed(data):
                yield self.pop()
            self.put(data)
        if not self.drop_last and self.buffer:
            yield self.pop()

    @staticmethod
    def collate_fn(batch):
        return batch[0]
    
    def encode_tags(self, token_ids, tag):
        tag_hash = self._get_tag_hash(tag)
        self.tag_to_hash[tag] = tag_hash
        return torch.full((len(token_ids),), tag_hash, dtype=torch.int64)
    
    def _get_tag_hash(self, tag: str) -> int:
        hash_obj = hashlib.sha256(tag.encode('utf-8'))
        return int(hash_obj.hexdigest(), 16) & ((1 << 63) - 1)

if __name__ == '__main__':
    context = DatasetContext(world_size=1, rank=0, num_workers=1)

    dataset_info_list = [
        DatasetInfo(
            path="human.BaseDoc.sstable",
            weight=1,
            max_epoch=1
        ),
        DatasetInfo(
            path="hot_chars.BaseDoc.sstable",
            weight=2,
            max_epoch=2
        )
    ]
    
    dataset_info_list = DatasetInfoList(dataset_info_list)
    
    w_dataset = WeightedDataset(
        context=context,
        dataset_info_list=dataset_info_list
    )
    
    dataset = BatchedDataset(
        context=context,
        weighted_dataset=w_dataset,
        batch_size=1,
        max_len=8192
    )
    
    for data in dataset:
        print(data)
        print(data['indexes'].keys())
        print(data['indexes'].values())
    


import torch
from modelbest_sdk.dataset.modelbest_dataloader import CpmFlashAttnDataloader
from modelbest_sdk.dataset.thrift_wrapper.dataset_context import DatasetContext
from test.test_base import TestBase


class TestBatchedDataset(TestBase):
    def test_pack_batch(self):
        context = DatasetContext.load_from_file(self.simple_dataset_context_path)
        dataloader = CpmFlashAttnDataloader(context, self.simple_dataset_info_list, batch_size=1, max_len=16, cuda_prefetch=False)
        for batch in dataloader:
            '''
            doc.token_ids = [-1, 2, 3, 4, 5, 6, 0]
            doc.mask = [False, False, False, False, False, False, True]
            '''
            assert (batch['input_ids'] == torch.tensor([[-1, 2, 3, 4, 5, 6, 0, -1, 2, 3, 4, 5, 6, 0, 0, 0]])).all()
            assert batch['input_ids'].shape == torch.Size([1, 16])
            assert (batch['target_ids'] == torch.tensor([[2, 3, 4, 5, 6, -100, -100, 2, 3, 4, 5, 6, -100, -100, -100, -100]])).all()
            assert batch['target_ids'].shape == torch.Size([1, 16])
            assert (batch['cu_seqlens'] == torch.tensor([ 0,  7, 14, 16])).all()
            assert (batch['position_ids'] == torch.tensor([[0, 1, 2, 3, 4, 5, 6, 0, 1, 2, 3, 4, 5, 6, 0, 0]])).all()


import os
import unittest
import torch.distributed as dist
import torch.multiprocessing as mp

from modelbest_sdk.dataset.modelbest_dataloader import CpmFlashAttnDataloader
from modelbest_sdk.dataset.thrift_wrapper.dataset_context import DatasetContext
from modelbest_sdk.dataset.thrift_wrapper.dataset_checkpoint import Chunk
from modelbest_sdk.dataset.thrift_wrapper.dataset_context import DatasetContext
from test.test_base import TestBase

class TestCheckpoint(TestBase):
    def test_simple_dataset_checkpoint(self):
        # dataset has 20 records, each records length is 7, max_epoch is 1
        # context suggest to use 16 as batch size * max len, so every 2 records will be grouped into one batch
        context = DatasetContext.load_from_file(self.simple_dataset_context_path)
        dataloader = CpmFlashAttnDataloader(context, self.simple_dataset_info_list, batch_size=1, max_len=16, cuda_prefetch=False)
        for i, batch in enumerate(dataloader, start=1):
            dataloader.update(batch['indexes'])
            if i == 5:
                break
        dataloader.save()
        assert dataloader.checkpoint().checkpoint_list[0].chunk_size == 16
        assert dataloader.checkpoint().checkpoint_list[0].num_chunks == 2
        
        context = DatasetContext.load_from_file(self.simple_dataset_context_path)
        dataloader = CpmFlashAttnDataloader(context, self.simple_dataset_info_list, batch_size=1, max_len=16, cuda_prefetch=False)
        dataloader.resume()

        for i, batch in enumerate(dataloader, start=1):
            dataloader.update(batch['indexes'])
            
        assert dataloader.checkpoint().checkpoint_list[0].used.active == {}
        assert dataloader.checkpoint().checkpoint_list[0].used.done == {0: {Chunk(epoch=0, start=0, stop=16), Chunk(epoch=0, start=16, stop=20)}}
    
    def test_dist_dataset_checkpoint(self):
        context = DatasetContext.load_from_file(self.dist_dataset_context_path)
        dataloader = CpmFlashAttnDataloader(context, self.dist_dataset_info_list, batch_size=1, max_len=16, cuda_prefetch=False)
        for i, batch in enumerate(dataloader, start=1):
            dataloader.update(batch['indexes'])
            # print(batch)
            if i == 30:
                break
        dataloader.save()
        assert dataloader.checkpoint().checkpoint_list[0].num_chunks == 3
        assert dataloader.checkpoint().checkpoint_list[0].chunk_size == 8
        
        dataloader = CpmFlashAttnDataloader(context, self.dist_dataset_info_list, batch_size=1, max_len=16, cuda_prefetch=False)
        dataloader.resume()
        for i, batch in enumerate(dataloader, start=1):
            dataloader.update(batch['indexes'])
            if i == 1001:
                break
        cnt_0 = cnt_1 = 0
        for chunk, set in dataloader.checkpoint().checkpoint_list[0].used.active.items():
            cnt_0 += len(set)
        for epoch, chunk_set in dataloader.checkpoint().checkpoint_list[0].used.done.items():
            for chunk in chunk_set:
                cnt_0 += chunk.stop - chunk.start
        for chunk, set in dataloader.checkpoint().checkpoint_list[1].used.active.items():
            cnt_1 += len(set)
        for epoch, chunk_set in dataloader.checkpoint().checkpoint_list[1].used.done.items():
            for chunk in chunk_set:
                cnt_1 += chunk.stop - chunk.start
        assert round(cnt_1 / cnt_0, 0) == 2 # weight_1 : weight_0 = 2 : 1
        
    
    def test_world_size_change(self):
        world_size = 2
        mp.spawn(process_run, args=(world_size, self.dist_dataset_context_path, self.dist_dataset_info_list, ), nprocs=world_size, join=True)
        world_size = 3
        mp.spawn(process_run, args=(world_size, self.dist_dataset_context_path, self.dist_dataset_info_list, ), nprocs=world_size, join=True)
        


def process_run(rank, world_size, context_path, dataset_info_list):
    print(f"rank {rank} start")
    print(f"world_size {world_size}")
    # 设置环境变量
    os.environ['MASTER_ADDR'] = 'localhost'
    os.environ['MASTER_PORT'] = '12355'
    
    # 初始化进程组
    dist.init_process_group("gloo", rank=rank, world_size=world_size)


    context = DatasetContext.load_from_file(context_path)
    context.rank = rank
    context.world_size = world_size
    dataloader = CpmFlashAttnDataloader(context, dataset_info_list, batch_size=1, max_len=16, cuda_prefetch=False)
    dataloader.resume()
    print(dataloader.checkpoint().checkpoint_list[0].used)
    for i, batch in enumerate(dataloader, start=1):
        dataloader.update(batch['indexes'])
        if i == 20:
            break
    dataloader.save()
    print(dataloader.checkpoint().checkpoint_list[0].used)
    
    dist.destroy_process_group()

if __name__ == '__main__':
    unittest.main()


from collections import defaultdict, deque
import time
import numpy as np
import torch

# from modelbest_sdk.dataset.collater.collater import Collater
from modelbest_sdk.dataset.weighted_dataset import WeightedDataset
from modelbest_sdk.dataset.thrift_wrapper.base_doc import BaseDoc
from modelbest_sdk.dataset.thrift_wrapper.dataset_checkpoint import *
from modelbest_sdk.dataset.thrift_wrapper.dataset_context import DatasetContext

def default_factory_list():
    return list()

def default_factory_dict():
    return defaultdict(default_factory_list)

class SegmentDataset(torch.utils.data.IterableDataset):
    def __init__(self, context, weighted_dataset, max_len, drop_last=False):
        self.context = context
        self.weighted_dataset = weighted_dataset
        self.max_len = max_len
        self.drop_last = drop_last
        
        self.buffer = deque()
        self.current_length = 0

        # 预计算 attention_mask
        self.attention_mask = self._create_attention_mask(max_len)
        
    def _create_attention_mask(self, max_len):
        mask = torch.tril(torch.ones((max_len, max_len))).unsqueeze(0) < 0.5
        return mask

    def __iter__(self):
        for data in self.weighted_dataset:
            while self.current_length > self.max_len:
                yield self.pop()
            self.put(data)
        while self.buffer:
            yield self.pop()
    
    def put(self, data):
        self.buffer.append(data)
        self.current_length += len(data['doc'].token_ids)
    
    def pop(self):
        tokens = torch.zeros((self.max_len + 1), dtype=torch.int64)
        mask = torch.zeros((self.max_len + 1), dtype=torch.bool)
        position_ids = torch.arange(self.max_len, dtype=torch.int64)
        indexes = defaultdict(default_factory_dict)

        remaining_seq_length = self.max_len + 1
        begin = 0
        while self.buffer and remaining_seq_length > 0:
            data = self.buffer.popleft()
            doc = data['doc']
            index = data['index']
            chunk = data['chunk']
            dataset_idx = data['dataset_idx']
            indexes[int(dataset_idx)][chunk].append(index)
            if remaining_seq_length > len(doc.token_ids):
                end = begin + len(doc.token_ids)
                doc_to_fill = doc
            else:
                end = begin + remaining_seq_length
                doc1, doc2 = doc.split(remaining_seq_length)
                doc_to_fill = doc1
                self.buffer.appendleft({'doc': doc2, 'chunk': data['chunk'], 'index': data['index'], 'dataset_idx': data['dataset_idx']})
                self.current_length += len(doc2.token_ids)
            
            tokens[begin:end] = torch.tensor(doc_to_fill.token_ids, dtype=torch.int64)
            mask[begin:end] = torch.tensor(doc_to_fill.mask, dtype=torch.bool)
            remaining_seq_length -= len(doc_to_fill.token_ids)
            self.current_length -= len(doc_to_fill.token_ids)
            begin = end

        if remaining_seq_length > 0 and self.drop_last:
            raise StopIteration

        return {
            'tokens': tokens[:-1],
            'labels': tokens[1:],
            'loss_mask': (~mask[:-1]).float(),
            'attention_mask': self.attention_mask,
            'position_ids': position_ids,
            'indexes': indexes
        }
    
    def collate_fn(batch):
        # 'batch' 是一个列表，其中包含所有 `__getitem__` 的返回值
        # 初始化一个字典来存储所有批处理后的数据
        batched_data = {k: [] for k in batch[0]}
        
        # 遍历批次中的每个样本
        for item in batch:
            for key in item:
                batched_data[key].append(item[key])

        # 对于可以直接转换为张量的数据，使用 torch.stack
        for key in ['tokens', 'labels', 'loss_mask', 'attention_mask', 'position_ids']:
            batched_data[key] = torch.stack(batched_data[key])

        # 对于 'indexes' 这样的复杂结构，保持原样或根据需要处理
        # 这里我们保持它作为一个列表
        batched_data['indexes'] = batched_data['indexes'][0]
        return batched_data
            
if __name__ == '__main__':
    context = DatasetContext(world_size=1, rank=0, num_workers=1)
    context.world_size = 8
    context.rank = 6
    dataset_info_list = [
        DatasetInfo(
            path="/home/emr-user/modelbest_sdk/example/mbtable_data/base_doc_simple",
            weight=1,
            max_epoch=1
        )
    ]
    
    dataset_info_list = DatasetInfoList(dataset_info_list)
    
    w_dataset = WeightedDataset(
        context=context,
        dataset_info_list=dataset_info_list
    )
    
    dataset = SegmentDataset(
        context=context,
        weighted_dataset=w_dataset,
        max_len=16
    )
    
    dataloader = torch.utils.data.DataLoader(
        dataset,
        batch_size=2,
        collate_fn=SegmentDataset.collate_fn
    )
    i = 0
    for data in dataloader:
        print(data)
    
